"""Functions to help clean & normalize business names.

See http://www.unicode.org/reports/tr15/#Normalization_Forms_Table for details
on Unicode normalization and the NFKD normalization used here.

Basic usage:

>> terms = prepare_default_terms()
>> basename("Daddy & Sons, Ltd.", terms, prefix=True, middle=True, suffix=True)
Daddy & Sons

"""

import functools
import operator
import re
import unicodedata

tail_removal_rexp = re.compile(r"[^\.\w]+$", flags=re.UNICODE)


terms_by_type = {
    "Corporation": [
        "company",
        "incorporated",
        "corporation",
        "corp.",
        "corp",
        "inc",
        "& co.",
        "& co",
        "inc.",
        "s.p.a.",
        "n.v.",
        "a.g.",
        "ag",
        "nuf",
        "s.a.",
        "s.f.",
        "oao",
        "co.",
        "co",
    ],
    "General Partnership": [
        "soc.col.",
        "stg",
        "d.n.o.",
        "ltda.",
        "v.o.s.",
        "a spol.",
        "ve\xc5\x99. obch. spol.",
        "kgaa",
        "o.e.",
        "s.f.",
        "s.n.c.",
        "s.a.p.a.",
        "j.t.d.",
        "v.o.f.",
        "sp.j.",
        "og",
        "sd",
        " i/s",
        "ay",
        "snc",
        "oe",
        "bt.",
        "s.s.",
        "mb",
        "ans",
        "da",
        "o.d.",
        "hb",
        "pt",
    ],
    "Joint Stock / Unlimited": [
        "unltd",
        "ultd",
        "sal",
        "unlimited",
        "saog",
        "saoc",
        "aj",
        "yoaj",
        "oaj",
        "akc. spol.",
        "a.s.",
    ],
    "Joint Venture": ["esv", "gie", "kv.", "qk"],
    "Limited": [
        "pty. ltd.",
        "pty ltd",
        "ltd",
        "l.t.d.",
        "bvba",
        "d.o.o.",
        "ltda",
        "gmbh",
        "g.m.b.h",
        "kft.",
        "kht.",
        "zrt.",
        "ehf.",
        "s.a.r.l.",
        "d.o.o.e.l.",
        "s. de r.l.",
        "b.v.",
        "tapui",
        "sp. z.o.o.",
        "sp. z o.o.",
        "spółka z o.o.",
        "s.r.l.",
        "s.l.",
        "s.l.n.e.",
        "ood",
        "oy",
        "rt.",
        "teo",
        "uab",
        "scs",
        "sprl",
        "limited",
        "bhd.",
        "sdn. bhd.",
        "sdn bhd",
        "as",
        "lda.",
        "tov",
        "pp",
    ],
    "Limited Liability Company": [
        "pllc",
        "llc",
        "l.l.c.",
        "plc.",
        "plc",
        "hf.",
        "oyj",
        "a.e.",
        "nyrt.",
        "p.l.c.",
        "sh.a.",
        "s.a.",
        "s.r.l.",
        "srl.",
        "srl",
        "aat",
        "3at",
        "d.d.",
        "s.r.o.",
        "spol. s r.o.",
        "s.m.b.a.",
        "smba",
        "sarl",
        "nv",
        "sa",
        "aps",
        "a/s",
        "p/s",
        "sae",
        "sasu",
        "eurl",
        "ae",
        "cpt",
        "as",
        "ab",
        "asa",
        "ooo",
        "dat",
        "vat",
        "zat",
        "mchj",
        "a.d.",
    ],
    "Limited Liability Limited Partnership": ["lllp", "l.l.l.p."],
    "Limited Liability Partnership": ["llp", "l.l.p.", "sp.p.", "s.c.a.", "s.c.s."],
    "Limited Partnership": [
        "gmbh & co. kg",
        "lp",
        "l.p.",
        "s.c.s.",
        "s.c.p.a",
        "comm.v",
        "k.d.",
        "k.d.a.",
        "s. en c.",
        "e.e.",
        "s.a.s.",
        "s. en c.",
        "c.v.",
        "s.k.a.",
        "sp.k.",
        "s.cra.",
        "ky",
        "scs",
        "kg",
        "kd",
        "k/s",
        "ee",
        "secs",
        "kda",
        "ks",
        "kb",
        "kt",
    ],
    "Mutual Fund": ["sicav"],
    "No Liability": ["nl"],
    "Non-Profit": ["vzw", "ses.", "gte."],
    "Private Company": ["private", "pte", "xk"],
    "Professional Corporation": ["p.c.", "vof", "snc"],
    "Professional Limited Liability Company": ["pllc", "p.l.l.c."],
    "Sole Proprietorship": [
        "e.u.",
        "s.p.",
        "t:mi",
        "tmi",
        "e.v.",
        "e.c.",
        "et",
        "obrt",
        "fie",
        "ij",
        "fop",
        "xt",
    ],
}

terms_by_country = {
    "Albania": ["sh.a.", "sh.p.k."],
    "Argentina": [
        "s.a.",
        "s.r.l.",
        "s.c.p.a",
        "scpa",
        "s.c.e i.",
        "s.e.",
        "s.g.r",
        "soc.col.",
    ],
    "Australia": ["nl", "pty. ltd.", "pty ltd"],
    "Austria": ["e.u.", "stg", "gesbr", "a.g.", "ag", "og", "kg"],
    "Belarus": ["aat", "3at"],
    "Belgium": [
        "esv",
        "vzw",
        "vof",
        "snc",
        "comm.v",
        "scs",
        "bvba",
        "sprl",
        "cvba",
        "cvoa",
        "sca",
        "sep",
        "gie",
    ],
    "Bosnia / Herzegovina": ["d.d.", "a.d.", "d.n.o.", "d.o.o.", "k.v.", "s.p."],
    "Brazil": ["ltda", "s.a.", "pllc", "ad", "adsitz", "ead", "et", "kd", "kda", "sd"],
    "Bulgaria": ["ad", "adsitz", "ead", "et", "kd", "kda", "sd"],
    "Cambodia": ["gp", "sm pte ltd.", "pte ltd.", "plc ltd.", "peec", "sp"],
    "Canada": ["gp", "lp", "sp"],
    "Chile": [
        "eirl",
        "s.a.",
        "sgr",
        "s.g.r.",
        "ltda",
        "s.p.a.",
        "sa",
        "s. en c.",
        "ltda.",
    ],
    "Columbia": ["s.a.", "e.u.", "s.a.s.", "suc. de descendants", "sca"],
    "Croatia": ["d.d.", "d.o.o.", "obrt"],
    "Czech Republic": [
        "a.s.",
        "akc. spol.",
        "s.r.o.",
        "spol. s r.o.",
        "v.o.s.",
        "ve\xc5\x99. obch. spol.",
        "a spol.",
        "k.s.",
        "kom. spol.",
        "kom. spol.",
    ],
    "Denmark": [
        "i/s",
        "a/s",
        "k/s",
        "p/s",
        "amba",
        "a.m.b.a.",
        "fmba",
        "f.m.b.a.",
        "smba",
        "s.m.b.a.",
        "g/s",
    ],
    "Dominican Republic": [
        "c. por a.",
        "cxa",
        "s.a.",
        "s.a.s.",
        "srl.",
        "srl",
        "eirl.",
        "sa",
        "sas",
    ],
    "Ecuador": ["s.a.", "c.a.", "sa", "ep"],
    "Egypt": ["sae"],
    "Estonia": ["fie", "oü", "as"],
    "Finland": ["t:mi", "tmi", "as oy", "as.oy", "ay", "ky", "oy", "oyj", "ok"],
    "France": [
        "sicav",
        "sarl",
        "sogepa",
        "ei",
        "eurl",
        "sasu",
        "fcp",
        "gie",
        "sep",
        "snc",
        "scs",
        "sca",
        "scop",
        "sem",
        "sas",
    ],
    "Germany": [
        "gmbh & co. kg",
        "e.g.",
        "e.v.",
        "gbr",
        "ohg",
        "partg",
        "kgaa",
        "gmbh",
        "g.m.b.h.",
        "ag",
        "mbh & co. kg",
    ],
    "Greece": [
        "a.e.",
        "ae",
        "e.e.",
        "ee",
        "epe",
        "e.p.e.",
        "mepe",
        "m.e.p.e.",
        "o.e.",
        "oe",
        "ovee",
        "o.v.e.e.",
    ],
    "Guatemala": ["s.a.", "sa"],
    "Haiti": ["sa"],
    "Hong Kong": ["ltd", "unltd", "ultd", "limited"],
    "Hungary": [
        "e.v.",
        "e.c.",
        "bt.",
        "kft.",
        "kht.",
        "kkt.",
        "k.v.",
        "zrt.",
        "nyrt",
        "ev",
        "ec",
        "rt.",
    ],
    "Iceland": ["ehf.", "hf.", "ohf.", "s.f.", "ses."],
    "India": ["pvt. ltd.", "ltd.", "psu", "pse"],
    "Indonesia": ["ud", "fa", "pt"],
    "Ireland": ["cpt", "teo"],
    "Israel": ["b.m.", "bm", "ltd", "limited"],
    "Italy": ["s.n.c.", "s.a.s.", "s.p.a.", "s.a.p.a.", "s.r.l.", "s.c.r.l.", "s.s."],
    "Japan": ["k.k.", "g.k.", "gk", "y.k."],
    "Latvia": ["as", "sia", "ik", "ps", "ks"],
    "Lebanon": ["sal"],
    "Lithuania": ["uab", "ab", "ij", "mb"],
    "Luxemborg": ["s.a.", "s.a.r.l.", "secs"],
    "Macedonia": ["d.o.o.", "d.o.o.e.l", "k.d.a.", "j.t.d.", "a.d.", "k.d."],
    "Malaysia": ["bhd.", "sdn. bhd."],
    "Mexico": ["s.a.", "s. de. r.l.", "s. en c.", "s.a.b.", "s.a.p.i.", "s.a. de c.v."],
    "Mongolia": ["xk", "xxk"],
    "Netherlands": ["v.o.f.", "c.v.", "b.v.", "n.v."],
    "New Zealand": ["tapui", "ltd", "limited"],
    "Nigeria": ["gte.", "plc", "ltd.", "ultd."],
    "Norway": [
        "asa",
        "as",
        "ans",
        "ba",
        "bl",
        "da",
        "etat",
        "fkf",
        "hf",
        "iks",
        "kf",
        "ks",
        "nuf",
        "rhf",
        "sf",
    ],
    "Oman": ["saog", "saoc"],
    "Pakistan": ["ltd.", "pvt. ltd.", "ltd", "limited"],
    "Peru": ["sa", "s.a.", "s.a.a."],
    "Philippines": [
        "coop.",
        "corp.",
        "corp",
        "ent.",
        "inc.",
        "inc",
        "llc",
        "l.l.c.",
        "ltd.",
    ],
    "Poland": [
        "p.p.",
        "s.k.a.",
        "sp.j.",
        "sp.k.",
        "sp.p.",
        "sp. z.o.o.",
        "s.c.",
        "s.a.",
    ],
    "Portugal": ["lda.", "crl", "s.a.", "s.f.", "sgps"],
    "Romania": ["s.c.a.", "s.c.s.", "s.n.c.", "s.r.l.", "o.n.g.", "s.a."],
    "Russia": ["ooo", "oao", "zao", "3ao", "пао", "оао", "ооо"],
    "Serbia": ["d.o.o.", "a.d.", "k.d.", "o.d."],
    "Singapore": [
        "bhd",
        "pte ltd",
        "sdn bhd",
        "llp",
        "l.l.p.",
        "ltd.",
        "pte",
        "pte. ltd.",
    ],
    "Slovenia": ["d.d.", "d.o.o.", "d.n.o.", "k.d.", "s.p."],
    "Slovakia": [
        "a.s.",
        "akc. spol.",
        "s.r.o.",
        "spol. s r.o.",
        "k.s.",
        "kom. spol.",
        "v.o.s.",
        "a spol.",
    ],
    "Spain": [
        "s.a.",
        "s.a.d.",
        "s.l.",
        "s.l.l.",
        "s.l.n.e.",
        "s.c.",
        "s.cra",
        "s.coop",
        "sal",
        "sccl",
    ],
    "Sweden": ["ab", "hb", "kb"],
    "Switzerland": ["ab", "sa", "gmbh", "g.m.b.h.", "sarl", "sagl"],
    "Turkey": ["koop."],
    "Ukraine": ["dat", "fop", "kt", "pt", "tdv", "tov", "pp", "vat", "zat", "at"],
    "United Kingdom": [
        "plc.",
        "plc",
        "cic",
        "cio",
        "l.l.p.",
        "llp",
        "l.p.",
        "lp",
        "ltd.",
        "ltd",
        "limited",
    ],
    "United States of America": [
        "llc",
        "inc.",
        "corporation",
        "incorporated",
        "company",
        "limited",
        "corp.",
        "inc.",
        "inc",
        "llp",
        "l.l.p.",
        "pllc",
        "and company",
        "& company",
        "inc",
        "inc.",
        "corp.",
        "corp",
        "ltd.",
        "ltd",
        "& co.",
        "& co",
        "co.",
        "co",
        "lp",
    ],
    "Uzbekistan": ["mchj", "qmj", "aj", "oaj", "yoaj", "xk", "xt", "ok", "uk", "qk"],
}

# from https://stackoverflow.com/a/51230541
NON_NFKD_MAP = {
    "\u0181": "B",
    "\u1d81": "d",
    "\u1d85": "l",
    "\u1d89": "r",
    "\u028b": "v",
    "\u1d8d": "x",
    "\u1d83": "g",
    "\u0191": "F",
    "\u0199": "k",
    "\u019d": "N",
    "\u0220": "N",
    "\u01a5": "p",
    "\u0224": "Z",
    "\u0126": "H",
    "\u01ad": "t",
    "\u01b5": "Z",
    "\u0234": "l",
    "\u023c": "c",
    "\u0240": "z",
    "\u0142": "l",
    "\u0244": "U",
    "\u2c60": "L",
    "\u0248": "J",
    "\ua74a": "O",
    "\u024c": "R",
    "\ua752": "P",
    "\ua756": "Q",
    "\ua75a": "R",
    "\ua75e": "V",
    "\u0260": "g",
    "\u01e5": "g",
    "\u2c64": "R",
    "\u0166": "T",
    "\u0268": "i",
    "\u2c66": "t",
    "\u026c": "l",
    "\u1d6e": "f",
    "\u1d87": "n",
    "\u1d72": "r",
    "\u2c74": "v",
    "\u1d76": "z",
    "\u2c78": "e",
    "\u027c": "r",
    "\u1eff": "y",
    "\ua741": "k",
    "\u0182": "B",
    "\u1d86": "m",
    "\u0288": "t",
    "\u018a": "D",
    "\u1d8e": "z",
    "\u0111": "d",
    "\u0290": "z",
    "\u0192": "f",
    "\u1d96": "i",
    "\u019a": "l",
    "\u019e": "n",
    "\u1d88": "p",
    "\u02a0": "q",
    "\u01ae": "T",
    "\u01b2": "V",
    "\u01b6": "z",
    "\u023b": "C",
    "\u023f": "s",
    "\u0141": "L",
    "\u0243": "B",
    "\ua745": "k",
    "\u0247": "e",
    "\ua749": "l",
    "\u024b": "q",
    "\ua74d": "o",
    "\u024f": "y",
    "\ua751": "p",
    "\u0253": "b",
    "\ua755": "p",
    "\u0257": "d",
    "\ua759": "q",
    "\xd8": "O",
    "\u2c63": "P",
    "\u2c67": "H",
    "\u026b": "l",
    "\u1d6d": "d",
    "\u1d71": "p",
    "\u0273": "n",
    "\u1d75": "t",
    "\u1d91": "d",
    "\xf8": "o",
    "\u2c7e": "S",
    "\u1d7d": "p",
    "\u2c7f": "Z",
    "\u0183": "b",
    "\u0187": "C",
    "\u1d80": "b",
    "\u0289": "u",
    "\u018b": "D",
    "\u1d8f": "a",
    "\u0291": "z",
    "\u0110": "D",
    "\u0193": "G",
    "\u1d82": "f",
    "\u0197": "I",
    "\u029d": "j",
    "\u019f": "O",
    "\u2c6c": "z",
    "\u01ab": "t",
    "\u01b3": "Y",
    "\u0236": "t",
    "\u023a": "A",
    "\u023e": "T",
    "\ua740": "K",
    "\u1d8a": "s",
    "\ua744": "K",
    "\u0246": "E",
    "\ua748": "L",
    "\ua74c": "O",
    "\u024e": "Y",
    "\ua750": "P",
    "\ua754": "P",
    "\u0256": "d",
    "\ua758": "Q",
    "\u2c62": "L",
    "\u0266": "h",
    "\u2c73": "w",
    "\u2c6a": "k",
    "\u1d6c": "b",
    "\u2c6e": "M",
    "\u1d70": "n",
    "\u0272": "n",
    "\u1d92": "e",
    "\u1d74": "s",
    "\u2c7a": "o",
    "\u2c6b": "Z",
    "\u027e": "r",
    "\u0180": "b",
    "\u0282": "s",
    "\u1d84": "k",
    "\u0188": "c",
    "\u018c": "d",
    "\ua742": "K",
    "\u1d99": "u",
    "\u0198": "K",
    "\u1d8c": "v",
    "\u0221": "d",
    "\u2c71": "v",
    "\u0225": "z",
    "\u01a4": "P",
    "\u0127": "h",
    "\u01ac": "T",
    "\u0235": "n",
    "\u01b4": "y",
    "\u2c72": "W",
    "\u023d": "L",
    "\ua743": "k",
    "\u0249": "j",
    "\ua74b": "o",
    "\u024d": "r",
    "\ua753": "p",
    "\u0255": "c",
    "\ua757": "q",
    "\u2c68": "h",
    "\ua75b": "r",
    "\ua75f": "v",
    "\u2c61": "l",
    "\u2c65": "a",
    "\u01e4": "G",
    "\u0167": "t",
    "\u2c69": "K",
    "\u026d": "l",
    "\u1d6f": "m",
    "\u0271": "m",
    "\u1d73": "r",
    "\u027d": "r",
    "\u1efe": "Y",
}


def get_unique_terms():
    "retrieve all unique terms from termdata definitions"
    ts = functools.reduce(operator.iconcat, terms_by_type.values(), [])
    cs = functools.reduce(operator.iconcat, terms_by_country.values(), [])
    return set(ts + cs)


def remove_accents(t):
    """based on https://stackoverflow.com/a/51230541"""
    nfkd_form = unicodedata.normalize("NFKD", t.casefold())
    return "".join(
        NON_NFKD_MAP[c] if c in NON_NFKD_MAP else c
        for part in nfkd_form
        for c in part
        if unicodedata.category(part) != "Mn"
    )


def strip_punct(t):
    return t.replace(".", "").replace(",", "").replace("-", "")


def normalize_terms(terms):
    "normalize terms"
    return (strip_punct(remove_accents(t)) for t in terms)


def strip_tail(name):
    "get rid of all trailing non-letter symbols except the dot"
    match = re.search(tail_removal_rexp, name)
    if match is not None:
        name = name[: match.span()[0]]
    return name


def normalized(text):
    "caseless Unicode normalization"
    return remove_accents(text)


def prepare_default_terms():
    "construct an optimized term structure for basename extraction"
    terms = get_unique_terms()
    nterms = normalize_terms(terms)
    ntermparts = (t.split() for t in nterms)
    # make sure that the result is deterministic, sort terms descending by number of tokens, ascending by names
    sntermparts = sorted(ntermparts, key=lambda x: (-len(x), x))
    return [(len(tp), tp) for tp in sntermparts]


def custom_basename(name, terms, suffix=True, prefix=False, middle=False, **kwargs):
    "return cleaned base version of the business name"

    name = strip_tail(name)
    nparts = name.split()
    nname = normalized(name)
    nnparts = list(map(strip_punct, nname.split()))
    nnsize = len(nnparts)

    if suffix:
        for termsize, termparts in terms:
            if nnparts[-termsize:] == termparts:
                del nnparts[-termsize:]
                del nparts[-termsize:]

    if prefix:
        for termsize, termparts in terms:
            if nnparts[:termsize] == termparts:
                del nnparts[:termsize]
                del nparts[:termsize]

    if middle:
        for termsize, termparts in terms:
            if termsize > 1:
                sizediff = nnsize - termsize
                if sizediff > 1:
                    for i in range(0, nnsize - termsize + 1):
                        if termparts == nnparts[i : i + termsize]:
                            del nnparts[i : i + termsize]
                            del nparts[i : i + termsize]
            else:
                if termparts[0] in nnparts[1:-1]:
                    idx = nnparts[1:-1].index(termparts[0])
                    del nnparts[idx + 1]
                    del nparts[idx + 1]

    return strip_tail(" ".join(nparts))


# convenience for most common use cases that don't parametrize base name extraction
basename = functools.partial(custom_basename, terms=prepare_default_terms())
